import asyncio
import os
import random
from datetime import datetime
from typing import Text, Optional, Dict
from unittest.mock import patch

import pytest
import time
from _pytest.tmpdir import TempdirFactory

import rasax.community.utils as rasa_x_utils
from rasax.community.database.model import Model
from rasax.community.services.model_service import ModelService
from tests.unit.conftest import insert_test_model

TEST_MODEL_NAME = "test model"
TEST_PROJECT = "default"
TEST_TAG = "test tag"


@pytest.fixture(scope="module")
def model_service(session):
    # remove models which were stored during previous tests
    session.query(Model).delete()
    session.commit()

    return ModelService("path", session, "production")


def test_save_model_data(model_service: ModelService):
    expected = insert_test_model(TEST_PROJECT, TEST_MODEL_NAME, model_service)

    actual = model_service.get_model_by_name(TEST_PROJECT, TEST_MODEL_NAME)

    assert actual["project"] == TEST_PROJECT
    assert actual["model"] == TEST_MODEL_NAME
    assert actual["tags"] == []
    assert actual["hash"] == expected["model_hash"]
    assert actual["version"] == expected["_version"]
    assert actual["trained_at"] == expected["training_time"]
    assert actual["path"] == expected["stored_path"]


async def test_get_models(model_service: ModelService):
    model_service.minimum_compatible_version = asyncio.coroutine(lambda: "1.0")
    with patch.object(
        model_service, "is_model_compatible", lambda _, model_version: True
    ):
        model_service._discover_models = asyncio.coroutine(lambda: True)
        _, number_models = await model_service.get_models(TEST_PROJECT)

        assert number_models == 1


async def test_get_models_with_offset_limit(
    model_service: ModelService, new_project: Text
):
    model_service.minimum_compatible_version = asyncio.coroutine(lambda: "1.0")
    with patch.object(
        model_service, "is_model_compatible", lambda _, model_version: True
    ):
        test_model1 = "test_model1"
        training_time = time.time()
        insert_test_model(
            new_project, test_model1, model_service, training_time=training_time
        )
        test_model2 = "test_model2"
        insert_test_model(
            new_project, test_model2, model_service, training_time=training_time
        )
        insert_test_model(
            new_project, "test_model0", model_service, training_time=training_time - 10
        )

        models, number_models = await model_service.get_models(
            new_project, limit=1, offset=None, discover_models=False
        )

        assert number_models == 3
        assert models[0]["model"] == test_model1

        models, number_models = await model_service.get_models(
            new_project, limit=1, offset=1, discover_models=False
        )
        assert len(models) == 1
        assert number_models == 3
        assert models[0]["model"] == test_model2


async def test_tag_model(model_service: ModelService):
    with patch.object(
        model_service, "is_model_compatible", lambda _, model_version: True
    ):
        await model_service.tag_model(TEST_PROJECT, TEST_MODEL_NAME, TEST_TAG)

        assert model_service.model_for_tag(
            TEST_PROJECT, TEST_TAG
        ) == model_service.get_model_by_name(TEST_PROJECT, TEST_MODEL_NAME)


async def test_tag_model_with_incompatible_version(
    model_service: ModelService, new_project: Text
):
    model_name = "test_tag_model_with_incompatible_version"
    insert_test_model(new_project, model_name, model_service, version="1.0")

    async def mocked_minimum_version():
        return "2.0"

    with patch.object(
        model_service, "minimum_compatible_version", mocked_minimum_version
    ):
        with pytest.raises(ValueError):
            await model_service.tag_model(new_project, model_name, TEST_TAG)


def test_delete_tag(model_service: ModelService):
    model_service.delete_tag(TEST_PROJECT, TEST_MODEL_NAME, TEST_TAG)

    # flush since following tests are in the same transaction
    model_service.commit()

    assert model_service.model_for_tag(TEST_PROJECT, TEST_TAG) is None
    assert model_service.get_model_by_name(TEST_PROJECT, TEST_MODEL_NAME)["tags"] == []


def test_delete_model(model_service: ModelService):
    model_service.delete_model(TEST_PROJECT, TEST_MODEL_NAME)

    assert model_service.get_model_by_name(TEST_PROJECT, TEST_MODEL_NAME) is None


async def test_delete_not_existing_models_from_db(model_service: ModelService):
    insert_test_model(TEST_PROJECT, TEST_MODEL_NAME, model_service)
    insert_test_model(TEST_PROJECT, "should be deleted", model_service)

    available = ["available", TEST_MODEL_NAME]
    model_service.delete_not_existing_models_from_db(TEST_PROJECT, available)

    with patch.object(
        model_service, "is_model_compatible", lambda _, model_version: True
    ):
        actual, number_models = await model_service.get_models(TEST_PROJECT)
        assert number_models == 1
        assert actual[0]["model"] == TEST_MODEL_NAME


def test_get_model_count(model_service: ModelService):
    original_count = model_service.get_model_count()
    num_models = random.randint(1, 10)
    for i in range(num_models):
        insert_test_model(TEST_PROJECT, f"get_model_count_test_model{i}", model_service)

    assert original_count + num_models == model_service.get_model_count()


def test_create_model_server_url():
    host = "http://rasa-x:5002"
    os.environ["RASA_X_HOST"] = host
    tag = "development"

    actual = ModelService.get_model_server_url(tag)

    assert actual == "http://rasa-x:5002/api/projects/default/models/tags/development"


def test_model_name_from_path(model_service: ModelService):
    model_path = "/some/path/20190918-132305.tar.gz"

    # noinspection PyProtectedMember
    assert model_service._model_name_from_path(model_path) == "20190918-132305"


def test_get_model_training_time_from_file(
    rasa_model_path: Text, model_service: ModelService
):
    metadata = model_service.get_model_metadata(rasa_model_path)
    training_time = model_service.get_model_training_time_from_file_as_str(
        metadata, rasa_model_path
    )
    # training time is valid timestamp
    assert datetime.strptime(training_time, "%Y%m%d-%H%M%S")


def test_get_model_training_time_as_unix_timestamp(
    rasa_model_path: Text, model_service: ModelService
):
    metadata = model_service.get_model_metadata(rasa_model_path)
    timestamp = model_service.get_training_time_as_unix_timestamp(
        metadata, rasa_model_path
    )
    # timestamp produces valid datetime object
    assert datetime.fromtimestamp(timestamp)


@pytest.mark.parametrize(
    "metadata", [None, {"some": "key"}, {"trained_at": "some-nonsense"}]
)
def test_get_model_training_time_as_unix_timestamp_exception_cases(
    tmpdir_factory: TempdirFactory,
    model_service: ModelService,
    metadata: Optional[Dict],
):
    from pathlib import Path

    temp_dir = tmpdir_factory.mktemp("model-dir")
    dummy_model_path = os.path.join(temp_dir, "my_model.tar.gz")

    # create dummy file
    Path(dummy_model_path).touch()

    file_modification_time = os.path.getctime(dummy_model_path)

    # timestamp should be the same as getctime() when metadata is invalid
    assert (
        model_service.get_training_time_as_unix_timestamp(metadata, dummy_model_path)
        == file_modification_time
    )


def test_store_trained_model_on_disk(
    rasa_model_path: Text, model_service: ModelService
):
    # open model as bytes
    model_content_temporary = rasa_x_utils.file_as_bytes(rasa_model_path)

    # noinspection PyProtectedMember
    permanent_model_path = model_service._store_trained_model_on_disk(
        model_content_temporary
    )

    # open model at permanent location and ensure contents are the same as before
    model_content_permanent = rasa_x_utils.file_as_bytes(permanent_model_path)

    assert model_content_temporary == model_content_permanent


async def test_save_trained_model(rasa_model_path: Text, model_service: ModelService):
    import rasa

    # load model content as bytes and save it
    model_content = rasa_x_utils.file_as_bytes(rasa_model_path)
    model_name = await model_service.save_trained_model(TEST_PROJECT, model_content)

    model_service.commit()

    # fetch model from database
    model = model_service.get_model_by_name(TEST_PROJECT, model_name)
    assert model

    metadata = model_service.get_model_metadata(model["path"])

    # model name reflects training time
    assert model_name == model_service.get_model_training_time_from_file_as_str(
        metadata, model["path"]
    )

    # model name is correct
    assert model["model"] == model_name

    # model version is correct
    assert model["version"] == rasa.__version__

    # training time is correct
    assert (
        model_service.get_training_time_as_unix_timestamp(metadata, model["path"])
        == model["trained_at"]
    )


async def test_save_existing_model_error(
    rasa_model_path: Text, model_service: ModelService
):
    # load model content as bytes and save it
    model_content = rasa_x_utils.file_as_bytes(rasa_model_path)
    _ = await model_service.save_trained_model(TEST_PROJECT, model_content)

    # saving the same model again should not work as it already exists
    with pytest.raises(FileExistsError):
        _ = await model_service.save_trained_model(TEST_PROJECT, model_content)
